-- Coding standards brought to you by the East India Trading Company in association with the Save the Camel Foundation
local addonName, addonTable = ...
assert(type(addonName) == "string", "Critic requires WoW 3.3")

local L = LibStub("AceLocale-3.0"):GetLocale(addonName)
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
local SM = LibStub("LibSharedMedia-3.0")
local playerClass, db
local LGT = LibStub("LibGroupTalents-1.0")

assert(AceGUIWidgetLSMlists, addonName.." requires AceGUI-3.0-SharedMediaWidgets")

local next, pairs, select, tonumber, strmatch, type = _G.next, _G.pairs, _G.select, _G.tonumber, _G.strmatch, _G.type
local floor, ceil, abs, tremove, tinsert = _G.floor, _G.ceil, _G.abs, _G.tremove, _G.tinsert

--[===[@debug@
local d
do
	_G[addonName.."DEBUG"] = true
	local z
	local function dd(str)
		local r, g, b = z%3, z%4, z%5
		z = z + 1
		return format("|cFF%02X%02X%02X%s|r", r / 4 * 255, g / 4 * 255, b / 4 * 255, str)
	end
	function d(fmt, ...)
		if (_G[addonName.."DEBUG"]) then
			z = 3
			fmt = fmt:gsub("(%%[sdqxf])", dd)
			ChatFrame1:AddMessage("|cFFFF8080BH:|r "..format(fmt, ...), 0.8, 0.8, 0.8)
		end
	end
end
--@end-debug@]===]

local new, del, copy, deepDel
do
	local list = setmetatable({},{__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			t = {...}
			return t
		end
	end
	function del(t)
		if (t) then
			wipe(t)
			t[''] = true
			t[''] = nil
			list[t] = true
		end
	end
	function deepDel(t)
		if (t) then
			for k,v in pairs(t) do
				if type(v) == "table" then
					deepDel(v)
				end
				t[k] = nil
			end
			t[''] = true
			t[''] = nil
			list[t] = true
		end
	end
end

local module = LibStub("AceAddon-3.0"):NewAddon(addonName, "AceConsole-3.0", "AceEvent-3.0")
--[===[@debug@
_G[addonName] = module
--@end-debug@]===]
module.version = tonumber(strmatch("$Revision: 11 $", "(%d+)")) or 0
module.version = max(module.version, tonumber(strmatch("11", "(%d+)")) or 0)

local LDB = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject(addonName, {
	type = "launcher",
	label = L["TITLE"],
	icon = select(3, GetSpellInfo(30919)),
})

-- OnClick
function LDB.OnClick(self, button)
	if (not IsModifierKeyDown()) then
		if (button == "RightButton") then
			module:OpenConfig()
		elseif (button == "LeftButton") then
			if (module.frame and module.frame:IsShown()) then
				module:HideFrame()
			else
				module:ShowFrame(1)
			end
		end
	end
end

-- OnTooltipShow
function LDB.OnTooltipShow(tt)
	tt:AddLine(module.title)
end

local function disableIfNotHistory()
	return db.mode ~= "history"
end

module.options = {
	type = 'group',
	order = 15,
	name = L["TITLE"],
	desc = L["Configuration"],
	handler = module,
	get = function(info) return db[info[#info]] end,
	set = function(info, value) db[info[#info]] = value module:ShowWhen(true) end,
	args = {
		General = {
			type = 'group',
			order = 1,
			name = L["General Settings"],
			desc = L["General Settings"],
			args = {
				show = {
					type = 'group',
					name = L["Show When"],
					desc = L["When to display"],
					guiInline = true,
					order = 100,
					get = function(info) return db.show[info[#info]] end,
					set = function(info, value)
						db.show[info[#info]] = value
						module:ShowWhen(true)
					end,
					args = {
						desc = {
							order = 0,
							type = "description",
							name = L["Indicate when to show the display."],
						},
  						always = {
							type = 'toggle',
							name = L["Always"],
							desc = L["Always show"],
							order = 10,
						},
						solo = {
							type = 'toggle',
							name = SOLO,
							desc = L["Show when solo"],
							disabled = function() return db.show.always end,
							order = 20,
						},
						party = {
							type = 'toggle',
							name = PARTY,
							desc = L["Show when in a party"],
							disabled = function() return db.show.always end,
							order = 30,
						},
						raid = {
							type = 'toggle',
							name = RAID,
							desc = L["Show when in a raid"],
							disabled = function() return db.show.always end,
							order = 40,
						},
					},
				},
				incombat = {
					type = 'group',
					name = L["In-Combat Options"],
					desc = L["In-Combat Options"],
					guiInline = true,
					order = 200,
					args = {
						desc = {
							order = 1,
							type = "description",
							name = L["In-combat specific options."],
						},
						inalpha = {
							type = "range",
							name = L["Alpha"],
							desc = L["Alpha level in-combat"],
							order = 10,
							min = 0,
							max = 1,
							step = 0.01,
						},
					},
				},
				oocombat = {
					type = 'group',
					name = L["Out-of-Combat Options"],
					desc = L["Out-of-Combat Options"],
					guiInline = true,
					order = 300,
					args = {
						desc = {
							order = 1,
							type = "description",
							name = L["Out-of-combat specific options."],
						},
						outalpha = {
							type = "range",
							name = L["Alpha"],
							desc = L["Alpha level out-of-combat"],
							order = 10,
							min = 0,
							max = 1,
							step = 0.01,
						},
					},
				},
				display = {
					type = 'group',
					name = L["Display Options"],
					desc = L["Display Options"],
					guiInline = true,
					order = 500,
					args = {
						desc = {
							order = 0,
							type = "description",
							name = L["The location of the display can be changed by holding down shift and dragging."],
						},
						mode = {
							type = 'select',
							name = L["Mode"],
							desc = L["Choose the type of display"],
							values = {history = "History", text = "Text"},
							order = 1,
						},
						height = {
							type = "range",
							name = L["Height"],
							desc = L["Height"],
							order = 10,
							min = 10,
							max = 100,
							step = 1,
						},
						width = {
							type = "range",
							name = L["Width"],
							desc = L["Width"],
							order = 20,
							min = 50,
							max = 500,
							step = 1,
							bigStep = 10,
						},
						scale = {
							type = "range",
							name = L["Scale"],
							desc = L["Scale"],
							order = 30,
							min = 0.2,
							max = 5,
							step = 0.01,
							bigStep = 0.1,
						},
						labels = {
							type = 'select',
							name = L["Label Placement"],
							desc = L["Which side to put the labels"],
							values = {LEFT = "Left", RIGHT = "Right"},
							order = 40,
						},
						duration = {
							type = "range",
							name = L["Duration"],
							desc = L["Set the duration in seconds that the bar is to represent"],
							order = 50,
							min = 5,
							max = 600,
							step = 5,
							disabled = disableIfNotHistory,
						},
						update = {
							type = "range",
							name = L["Update Frequence"],
							desc = L["How many times per second to update the display"],
							order = 60,
							min = 1,
							max = 60,
							step = 1,
							disabled = disableIfNotHistory,
						},
						mouseoverLabels = {
							type = "toggle",
							name = L["Mouseover Labels"].." (NYI)",
							desc = L["Only show labels when mousing over"],
							order = 70,
						},
					},
				},
				output = {
					type = 'group',
					name = L["Output Options"],
					desc = L["Output Options"],
					guiInline = true,
					order = 500,
					args = {
						warncap = {
							type = "toggle",
							name = L["Chat Warnings"],
							desc = L["Show warnings in chat log when you are crit capped"],
							order = 10,
						},
						summary = {
							type = "toggle",
							name = L["Summary Warning"],
							desc = L["Show warning at end of combat if you were crit capped and for how much of the fight"],
							order = 10,
						},
					},
				},
			},
		},
	},
}
AceConfig:RegisterOptionsTable(addonName, module.options)

-- SmoothColour
local function SmoothColour(percentage, light)
	local r, g
	if (percentage < 0.5) then
		g = min(1, max(0, 2*percentage))
		r = 1
		if (light) then
			g = g / 2 + 0.5
		end
	else
		g = 1
		r = min(1, max(0, 2*(1 - percentage)))
		if (light) then
			r = r / 2 + 0.5
		end
	end
	return r, g, light and 0.5 or 0
end

do
	local slots = {}

	local function scrollerCreateSlot(self)
		return self:CreateTexture(nil, "BACKGROUND")
	end

	-- scroller.GetSlot
	local function scrollerGetSlot(self)
		local slot = tremove(slots) or self:CreateSlot()
		slot:SetParent(self)
		slot:Show()
		slot:SetHeight(self:GetHeight())
		slot.startTime = GetTime()
		slot.endTime = nil
		return slot
	end

	-- scroller.FreeSlot
	local function scrollerFreeSlot(self, slot)
		slot:SetParent(module.frame)
		slot:Hide()
		slot:ClearAllPoints()
		tinsert(slots, slot)
	end

	-- scroller.OnUpdate
	local function scrollerOnUpdate(self, elapsed)
		local now = GetTime()
		if (not self.lastTick or now >= self.lastTick + (1 / db.update)) then
			self.lastTick = now

			local dur = db.duration
			local barStartTime = now - dur
			local kill = 0
			local prev
			local range = self.max - self.min

			for i,slot in ipairs(self.list) do
				if (slot.endTime and slot.endTime < now - dur) then
					slot:Hide()
					kill = kill + 1
				else
					slot:ClearAllPoints()
					if (slot.startTime <= barStartTime) then
						slot:SetPoint("BOTTOMLEFT")
					else
						if (not prev) then
							slot:SetPoint("BOTTOMLEFT", db.width / db.duration * (slot.startTime - barStartTime), 0)
						else
							slot:SetPoint("BOTTOMLEFT", prev, "BOTTOMRIGHT")
						end
					end

					if (slot.endTime) then
						if (slot.startTime <= barStartTime) then
							slot:SetWidth(db.width / db.duration * (slot.endTime - barStartTime))
						else
							slot:SetWidth(db.width / db.duration * (slot.endTime - slot.startTime))
						end
					else
						slot:SetPoint("BOTTOMRIGHT")
					end

					if (slot.val > slot.cap) then
						slot:SetTexture(1, 0, 0)
					elseif (slot.val > slot.cap - 3) then
						local r, g, b = SmoothColour((slot.cap - slot.val) / 3)
						slot:SetTexture(r, g, b)
					else
						slot:SetTexture(0, 1, 0)
					end

					slot:SetHeight(db.height / range * (slot.val - self.min))
				end

				prev = slot
			end

			for i = 1,kill do
				local slot = tremove(self.list, 1)
				if (slot) then
					self:FreeSlot(slot)
				end
			end
		end
	end

	-- scroller.Reset
	local function scrollerReset(self)
		while true do
			local slot = tremove(self.list)
			if (not slot) then
				break
			end
			self:FreeSlot(slot)
		end
	end

	-- scroller.AddTail
	local function scrollerAddTail(self, val, cap)
		local prevSlot = self.list[#self.list]
		if (prevSlot and val == prevSlot.val) then
			return
		end

		local slot = self:GetSlot()
		slot.val = val
		slot.cap = cap
		tinsert(self.list, slot)

		if (prevSlot) then
			prevSlot.endTime = slot.startTime
		end

		local range = self.max - self.min
		local capy = db.height / range * (cap - self.min)
		self.capbar:ClearAllPoints()
		self.capbar:SetPoint("LEFT", self, "BOTTOMLEFT", 0, capy)
		self.capbar:SetWidth(db.width)

		self.lastTick = nil
	end

	-- scroller.SetMinMaxValues
	local function scrollerSetMinMaxValues(self, n, x)
		self.min = n
		self.max = x
	end

	function module:CreateScroller(parent)
		local s = CreateFrame("Frame", nil, parent)
		s.list = {}

		s:SetScript("OnUpdate", scrollerOnUpdate)

		s.Reset				= scrollerReset
		s.SetMinMaxValues	= scrollerSetMinMaxValues
		s.GetSlot			= scrollerGetSlot
		s.FreeSlot			= scrollerFreeSlot
		s.CreateSlot		= scrollerCreateSlot
		s.AddTail			= scrollerAddTail

		s.capbar = s:CreateTexture(nil, "OVERLAY")
		s.capbar:SetTexture(1, 0, 0)
		s.capbar:SetHeight(0.4)

		return s
	end
end

do
	-- frame.OnDragStart
	local function frameOnShow(self)
		self:RegisterEvent("MODIFIER_STATE_CHANGED")
		self:Refresh()
	end

	-- frame.OnDragStop
	local function frameOnHide(self)
		self:UnregisterAllEvents()
	end

	-- frame.OnDragStop
	local function frameOnDragStop(self)
		self:StopMovingOrSizing()
		module:SavePosition(self)
	end

	-- frameOnMouseWheel
	local function frameOnMouseWheel(self, delta)
		if (IsShiftKeyDown()) then
			local scale = self:GetScale()
			self:SetScale(scale + (delta / 10))
			db.scale = scale
		end
	end

	-- frame.OnEvent
	local function frameOnEvent(self, event, arg1, arg2)
		if (event == "MODIFIER_STATE_CHANGED") then
			if (arg1 == "LSHIFT" or arg1 == "RSHIFT") then
				if (arg2 == 1) then
					self:EnableMouse(true)
					self:EnableMouseWheel(true)
				else
					frameOnDragStop(self)
					self:EnableMouse(false)
					self:EnableMouseWheel(false)
				end
			end
		end
	end

	-- frame.OnUpdate
	local function frameOnUpdate(self, elapsed)
		local a = self:GetAlpha()
		if (self.targetAlpha < a) then
			local a = max(0, self:GetAlpha() - elapsed * 2)
			if (a <= self.targetAlpha) then
				if (a <= 0) then
					self:Hide()
				end
				a = self.targetAlpha
			end
			self:SetAlpha(a)

		elseif (self.targetAlpha > a) then
			local a = self:GetAlpha() + elapsed * 2
			if (a >= self.targetAlpha) then
				a = self.targetAlpha
			end
			self:SetAlpha(a)
		end
	end

	-- frame.Open
	local function frameOpen(self, targetAlpha)
		self.targetAlpha = targetAlpha
		if (not self:IsShown()) then
			self:Show()
			self:Refresh()
		end
	end

	-- frame.Close
	local function frameClose(self)
		if (self:IsShown()) then
			self.targetAlpha = 0
		end
	end

	-- frame.Refresh
	local function frameRefresh(self)
		module:DoStats()

		local critLow = floor(module.critLow)
		local critHigh = ceil(max(module.critHigh, module.critCap))
		self.scroller:SetMinMaxValues(critLow, critHigh)

		if (module.critChance ~= self.lastCritChance or module.critCap ~= self.lastCritCap) then
			self.lastCritChance = module.critChance
			self.lastCritCap = module.critCap

			if (module.critChance > module.critCap) then
				if (UnitAffectingCombat("player")) then
					self.cappedStart = GetTime()
					if (db.warncap) then
						module:Print(format(L["|cFFFF0000Crit Capped! |cFFFF8080%.2f|r of |cFF80FF80%.2f|r max (|cFFFF8080%+.2f|r too much crit)"], module.critChance, module.critCap, module.critChance - module.critCap))
					end
				else
					self.cappedStart = nil
				end

			elseif (self.cappedStart) then
				if (UnitAffectingCombat("player")) then
					module.cappedTime = module.cappedTime + (GetTime() - self.cappedStart)
					self.cappedStart = GetTime()
				end
			end

			local r, g, b
			if (module.critChance > module.critCap) then
				r, g, b = 1, 0, 0
			elseif (module.critChance > module.critCap - 3) then
				r, g, b = SmoothColour((module.critCap - module.critChance) / 3)
			else
				r, g, b = 0, 1, 0
			end

			if (db.mode == "history") then
				self.scroller:AddTail(module.critChance, module.critCap)

				self.text:SetFormattedText("%.2f%% (%+.2f%%)", module.critChance, module.critChance - module.critCap)
				self.text:SetTextColor(r, g, b)

				self.capText:SetFormattedText("%.2f%%", module.critCap)
				self.lowText:SetFormattedText("%2d%%", floor(module.critLow))

			elseif (db.mode == "text") then
				local colour = format("|cFF%02X%02X%02X", r * 255, g * 255, b * 255)

				self.text:SetFormattedText("Crit: %s%.2f%%|r\nCap: |cFFA0A0A0%.2f%%|r\nSpare: %s%+.2f%%|r",
					colour, module.critChance,
					module.critCap,
					colour, module.critChance - module.critCap
				)
			end
		end
	end

	-- CreateDisplay
	function module:CreateDisplay()
		local f = CreateFrame("Frame", addonName.."Frame", UIParent)
		self.frame = f
		f:Hide()
		f:SetWidth(db.width or 200)
		f:SetHeight(db.height or 50)

		f:SetScale(db.scale or 1)

		f:SetMovable(true)
		f:RegisterForDrag("LeftButton")

		f:SetScript("OnShow",		frameOnShow)
		f:SetScript("OnHide",		frameOnHide)
		f:SetScript("OnDragStart",	f.StartMoving)
		f:SetScript("OnDragStop",	frameOnDragStop)
		f:SetScript("OnMouseWheel", frameOnMouseWheel)
		f:SetScript("OnEvent",		frameOnEvent)
		f:SetScript("OnUpdate",		frameOnUpdate)

		f.Open		= frameOpen
		f.Close		= frameClose
		f.Refresh	= frameRefresh
		f.Reset		= frameReset

		f.bg = f:CreateTexture(nil, "BACKGROUND")
		f.bg:SetTexture(0, 0, 0, 0.5)
		f.bg:SetAllPoints()

		f.scroller = self:CreateScroller(f)
		f.scroller:SetAllPoints()

		f.text = f.scroller:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		f.capText = f:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		f.lowText = f:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")

		module:RestorePosition(f)

		self.CreateDisplay = nil
	end
end

-- ShowFrame
function module:ShowFrame(targetAlpha)
	if (not self.frame) then
		self:CreateDisplay()
	end
	self.frame:SetScale(db.scale)
	self.frame:SetWidth(db.width)
	self.frame:SetHeight(db.height)

	local notSide = db.labels == "RIGHT" and "LEFT" or "RIGHT"

	self.frame.text:ClearAllPoints()
	self.frame.capText:ClearAllPoints()
	self.frame.lowText:ClearAllPoints()
	self.frame.text:SetJustifyH(db.labels)

	if (db.mode == "history") then
		self.frame.text:SetPoint(db.labels, 0, 3)
		self.frame.text:SetParent(self.frame.scroller)
		self.frame.capText:SetPoint("TOP"..notSide, self.frame, "TOP"..db.labels, 0, 3)
		self.frame.lowText:SetPoint("BOTTOM"..notSide, self.frame, "BOTTOM"..db.labels, 0, 1)
		self.frame.capText:Show()
		self.frame.lowText:Show()
		self.frame.scroller:Show()
		self.frame.capText:SetJustifyH(notSide)
		self.frame.lowText:SetJustifyH(notSide)

	elseif (db.mode == "text") then
		self.frame.text:SetPoint("TOP"..db.labels)
		self.frame.text:SetParent(self.frame)
		self.frame.capText:Hide()
		self.frame.lowText:Hide()
		self.frame.scroller:Hide()
	end

	self.frame:Open(targetAlpha)
end

-- HideFrame
function module:HideFrame()
	if (self.frame) then
		self.frame:Close()
	end
end

-- ShowWhen
function module:ShowWhen(reset)
	local role = LGT:GetUnitRole("player")
	if (role ~= "melee" and role ~= "tank") then
		self:HideFrame()
		return
	end

	local wantToShow
	if (db.show.always) then
		wantToShow = true
	else
		if (db.show.raid and GetNumRaidMembers() > 0) then
			wantToShow = true

		elseif (db.show.party and GetNumPartyMembers() > 0 and GetNumRaidMembers() == 0) then
			wantToShow = true

		elseif (db.show.solo and GetNumPartyMembers() == 0 and GetNumRaidMembers() == 0) then
			wantToShow = true
		end
	end

	if (wantToShow) then
		if (UnitAffectingCombat("player")) then
			wantToShow = db.inalpha
		else
			wantToShow = db.outalpha
		end
	end

	if (wantToShow and wantToShow > 0) then
		self:ShowFrame(wantToShow)
		if (reset) then
			self.frame.lastCritChance = nil
			self.frame:Refresh()
		end
		return
	end

	self:HideFrame()
end

-- SavePosition
function module:SavePosition(frame)
	local point, _, relPoint, x, y = frame:GetPoint(1)
	if (point) then
		if (not self.db.char.pos) then
			self.db.char.pos = {}
		end
		self.db.char.pos[frame:GetName()] = {point, "UIParent", relPoint, x, y}
		if (frame:IsResizable()) then
			if (not self.db.char.frameSizes) then
				self.db.char.frameSizes = {}
			end
			self.db.char.frameSizes[frame:GetName()] = {frame:GetWidth(), frame:GetHeight()}
		end
	end
end

-- RestorePosition
function module:RestorePosition(frame)
	local pos = self.db.char.pos and self.db.char.pos[frame:GetName()]
	if (pos) then
		frame:SetPoint(unpack(pos))
		if (self.db.char.frameSizes) then
			if (frame:IsResizable()) then
				local size = self.db.char.frameSizes[frame:GetName()]
				if (size) then
					frame:SetWidth(size[1])
					frame:SetHeight(size[2])
				end
			else
				self.db.char.frameSizes[frame:GetName()] = nil
			end
		end
	else
		frame:SetPoint("CENTER")
	end
end

-- PLAYER_LEAVING_WORLD
function module:PLAYER_LEAVING_WORLD()
	self:ShowWhen()
end
module.PLAYER_ENTERING_WORLD	= module.PLAYER_LEAVING_WORLD
module.RAID_ROSTER_UPDATE		= module.PLAYER_LEAVING_WORLD
module.PARTY_MEMBERS_CHANGED	= module.PLAYER_LEAVING_WORLD

function module:PLAYER_REGEN_DISABLED()
	self.cappedTime = 0
	self.startCombat = GetTime()
	self:ShowWhen()
end

-- PLAYER_REGEN_ENABLED
function module:PLAYER_REGEN_ENABLED()
	self:ShowWhen()
	self.endCombat = GetTime()
	if (db.summary) then
		if (self.startCombat and self.cappedTime and self.cappedTime > 0) then
			local dur = date("%X", self.startCombat - self.endCombat)
			if (dur) then
				self:Print(format(L["Combat Duration: |cFFFFFF80%s|r"], dur))
				self:Print(format(L["Crit Capped for: |cFFFFFF80%.1f%%|r (|cFFFFFF80%s|r)"], self.cappedTime / (self.startCombat - self.endCombat) * 100, date("%X", self.cappedTime)))
			end
		end
	end
	self.startCombat = nil
	self.cappedTime = 0
end

-- GetTargetModifiers
function module:GetTargetModifiers()
	self.mobCritChanceTaken = 0

	if (UnitExists("target")) then
		local name, rank, buff, count, debuffType, dur, max, caster = UnitDebuff("target", (GetSpellInfo(30706)))
		if (name) then
			-- Totem of Wrath
			self.mobCritChanceTaken = 3
		else
			local name, rank, buff, count, debuffType, dur, max, caster = UnitDebuff("target", (GetSpellInfo(54499)))
			if (name) then
				-- Heart of the Crusader
				local talented = LGT:UnitHasTalent(caster, (GetSpellInfo(20337)))		-- Yes, talent name can be different to debuff name (Russian locale for example)
				self.mobCritChanceTaken = (talented or 3)
			else
				-- TODO - Master Poisoner
			end
		end
	end
end

-- GetPlayerModifiers
function module:GetPlayerModifiers()
	-- Mostly hit modifiers, since these aren't reflected in tooltip data, where crit is
	local hitModifier = 0

	-- Heroic Presence (Draenei - Racial)
	if (UnitBuff("player", (GetSpellInfo(6562)))) then
		hitModifier = hitModifier + 1
	end

	if (playerClass == "ROGUE") then
		-- Precision (Rogue - Combat)
		local precision = LGT:UnitHasTalent("player", (GetSpellInfo(13705)))
		if (precision) then
			hitModifier = hitModifier + precision
		end

	elseif (playerClass == "WARRIOR") then
		-- Precision (Warrior - Fury)
		local precision = LGT:UnitHasTalent("player", (GetSpellInfo(29590)))
		if (precision) then
			hitModifier = hitModifier + precision
		end

	elseif (playerClass == "DEATHKNIGHT") then
		if (OffhandHasWeapon()) then
			-- Nerves of Cold Steel (DK - Frost)
			local nervesOfColdSteel = LGT:UnitHasTalent("player", (GetSpellInfo(49226)))
			if (nervesOfColdSteel) then
				hitModifier = hitModifier + nervesOfColdSteel
			end
		end

	elseif (playerClass == "SHAMAN") then
		if (OffhandHasWeapon()) then
			-- Dual Wield Specialization (Shaman - Enhancement)
			local dualWieldSpecialization = LGT:UnitHasTalent("player", (GetSpellInfo(30819)))
			if (dualWieldSpecialization) then
				hitModifier = hitModifier + dualWieldSpecialization * 2
			end
		end
	end

	self.playerHitModifier = hitModifier
end

-- DoStats
function module:DoStats()
	if (GetCritChance() <= 0) then
		-- Zoning, logging in etc
		return
	end

	self:GetPlayerModifiers()
	self:GetTargetModifiers()

	local dualWielding = OffhandHasWeapon()

	local playerLevel = UnitLevel("player")
	local targetLevel, isBoss
	if (UnitExists("target") and UnitCanAttack("target", "player")) then
		targetLevel = UnitLevel("target")
		isBoss = UnitClassification("target") == "worldboss"
	else
		targetLevel, isBoss = -1, true
	end
	if (isBoss) then
		targetLevel = UnitLevel("player") + 3
	end
	local levelDifference = targetLevel - playerLevel

	local critRating = GetCombatRating(CR_CRIT_MELEE)
	self.critChance = GetCritChance() + self.mobCritChanceTaken

	local hitRating = GetCombatRating(CR_HIT_MELEE)
	self.hitChance = GetCombatRatingBonus(CR_HIT_MELEE) + self.playerHitModifier

	local expertiseRating = GetCombatRating(CR_EXPERTISE)
	self.expertise = GetCombatRatingBonus(CR_EXPERTISE)

	local baseMissChance = dualWielding and 24 or 5
	self.missChance = max(0, min(100, baseMissChance + levelDifference - self.hitChance))

	local weaponSkill = UnitAttackBothHands("player")
	local mobDefense = 5 * targetLevel
	-- Accepted glance chance is 10 + defence - skill, but then we also know
	-- for a fact that glance chance vs a boss is 24%. so, using 9 for now
	self.glanceChance = max(0, min(100, 9 + mobDefense - weaponSkill))

	self.dodgeChance = max(0, min(100, (5 + (0.5 * levelDifference)) - self.expertise))

	self.critCap = 100 - self.missChance - self.dodgeChance - self.glanceChance

	if (not self.critLow or self.critChance < self.critLow) then
		self.critLow = self.critChance
	end
	if (not self.critHigh or self.critChance > self.critHigh) then
		self.critHigh = self.critChance
	end
end

-- UNIT_STATS
function module:UNIT_STATS(e, unit)
	if (unit == "player") then
		self:PLAYER_TARGET_CHANGED()
	end
end

-- PLAYER_TARGET_CHANGED
function module:PLAYER_TARGET_CHANGED()
	if (self.frame and self.frame:IsShown()) then
		self.frame:Refresh()
	end
end

-- ChangeAlphaOption
function module:ChangeAlphaOption(info, val)
	db.alpha = max(0.1, min(1, val))
	if (self.frame) then
		self.frame:SetAlpha(db.alpha)
	end
end

-- OpenConfig
function module:OpenConfig(input)
	InterfaceOptionsFrameAddOns.scrollFrame:SetVerticalScroll(0)
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames[addonName])
end

-- SetupOptions
function module:SetupOptions()
	self.optionsFrames = {}
	-- setup options table
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable(addonName, self.options)
	local ACD3 = LibStub("AceConfigDialog-3.0")

	self.optionsFrames[addonName] = ACD3:AddToBlizOptions(addonName, nil, nil, "General")
	self:RegisterChatCommand("bleedinghearts", "OpenConfig")

	self:RegisterModuleOptions("Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db), L["Profiles"])
	-- Add ordering data to the option table generated by AceDBOptions-3.0
	self.options.args.Profiles.order = -2

	self.SetupOptions = nil
end

-- RegisterModuleOptions
function module:RegisterModuleOptions(name, optionTbl, displayName)
	self.options.args[name] = (type(optionTbl) == "function") and optionTbl() or optionTbl
	self.optionsFrames[name] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName, displayName, addonName, name)
end

-- OnProfileChanged
function module:OnProfileChanged(event, database, newProfileKey)
	db = database.profile
end

-- OnModuleInitialize
function module:OnInitialize()
	playerClass = select(2, UnitClass("player"))

	self.db = LibStub("AceDB-3.0"):New("CriticDB", defaults, "Default")
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")

	self.db:RegisterDefaults({
		profile = {
			scale = 1,
			inalpha = 0.8,
			outalpha = 0,
			width = 200,
			height = 50,
			duration = 60,
			labels = "RIGHT",
			update = 1,
			mode = "history",
			summary = true,
			show = {
				always = false,
				combat = true,
				solo = false,
				party = false,
				raid = true,
			},
		}
	})
	db = self.db.profile

	self.callbacks = LibStub("CallbackHandler-1.0"):New(self)

	self.OnInitialize = nil
end

-- OnModuleEnable
function module:OnEnable()
	if (self.db) then
		self.title = L["TITLE"].." r"..module.version

		local class = select(2, UnitClass("player"))
		if (class ~= playerClass and self.OnModuleInitialize) then
			self:OnModuleInitialize()
		end

		playerClass = select(2, UnitClass("player"))

		self:SetupOptions()

		self:RegisterEvent("PLAYER_LEAVING_WORLD")
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("RAID_ROSTER_UPDATE")
		self:RegisterEvent("PARTY_MEMBERS_CHANGED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("UNIT_STATS")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")

		self:ShowWhen()
	end
end
